@@ -33,6 +33,11 @@
 (require 'dired)
 (require 'ert)
 
+(defvar bash-completion_test-notbash-prog (executable-find "dash")
+  "Path to a command line executable that is not bash.
+
+It must exit when given C-d as input.")
+
 (defvar bash-completion_test-setup-completion "/etc/bash_completion")
 (defconst bash-completion_test-start-mark "====START====")
 
@@ -47,6 +52,10 @@
             (list "--noediting"
                   "--noprofile"
                   "--rcfile" (expand-file-name "bashrc" test-env-dir)))
+           (bash-completion-short-command-timeout 3.0)
+           (bash-completion-command-timeout 10.0)
+           (bash-completion-initial-timeout 10.0)
+           (bash-completion-process-timeout 5.0)
            (completion-ignore-case nil)
            (completion-in-region-function 'completion--in-region)
            (completion-cycle-threshold 20)
@@ -56,6 +65,7 @@
                                (concat "explicit-"
                                        (file-name-nondirectory bash-completion-prog)
                                        "-args")))
+           (inhibit-field-text-motion nil)
            (old-explicit-args)
            (shell-mode-hook nil)
            (comint-mode-hook nil)
@@ -104,10 +114,6 @@
            (kill-process (get-buffer-process shell-buffer)))
          (kill-buffer shell-buffer)))))
 
-(defun bash-completion_test-bash-major-version ()
-  "Return the major version of the bash process."
-  (process-get (bash-completion--get-process) 'bash-major-version))
-
 (defun bash-completion_test-complete (complete-me)
   "Complete COMPLETE-ME and returns the resulting string."
   (goto-char (point-max))
@@ -117,19 +123,34 @@
   (buffer-substring-no-properties
    (line-beginning-position) (point)))
 
+(defun bash-completion_test-send-nowait (command)
+  "Execute COMMAND in the current shell buffer.
+
+This variant doesn't wait for the prompt to reappear."
+  (delete-region (line-beginning-position) (line-end-position))
+  (insert command)
+  (comint-send-input))
+
 (defun bash-completion_test-send (command &optional complete)
-  "Execute COMMAND in a shell buffer."
+  "Execute COMMAND in a shell buffer.
+
+Return a marker pointing to the beginning of the command."
   (goto-char (point-max))
+  (delete-region (line-beginning-position) (line-end-position))
   (let ((command-start (point)))
-    (delete-region (line-beginning-position) (line-end-position))
-    (insert command)
-    (when complete (completion-at-point))
-    (comint-send-input)
-    (bash-completion_test-wait-for-prompt command-start)))
+    (prog1 (point-marker)
+      (insert command)
+      (when complete (completion-at-point))
+      (comint-send-input)
+      (bash-completion_test-wait-for-prompt command-start))))
 
 (defun bash-completion_test-wait-for-prompt (&optional limit)
-  (bash-completion--wait-for-regexp
-   (get-buffer-process shell-buffer) "^.*$ " 3.0 limit))
+  (let ((process (get-buffer-process shell-buffer))
+        (no-timeout t))
+    (while (and no-timeout
+                (not (re-search-backward "^.*$ " limit t)))
+      (setq no-timeout (accept-process-output process 3.0 nil t)))
+    no-timeout))
 
 (defun bash-completion_test-buffer-string (&optional start end)
   (delete-trailing-whitespace (point-min) (point-max))
@@ -164,22 +185,26 @@ for testing completion."
       (with-temp-file (expand-file-name "bashrc" test-env-dir)
         ;; Disable ZSH warning under MacOS Catalina
         (insert "export BASH_SILENCE_DEPRECATION_WARNING=1") 
-        (insert "export PATH=/bin\n")
         (insert (format "cd '%s'\n" test-env-dir))
         (insert bashrc)
         (insert "\n")
         (insert "HISTFILE=/dev/null\n")
         (insert "history -c\n"))
       (let ((default-directory test-env-dir))
+        (with-temp-file "testfile1" (insert "test"))
+        (with-temp-file "testfile2" (insert "test"))
+        (with-temp-file "moretestfile" (insert "test"))
         (make-directory "some/directory" 'parents)
-        (make-directory "some/other/directory" 'parents)))))
+        (with-temp-file "some/testfile" (insert "test"))
+        (make-directory "some/other/directory" 'parents)
+        (with-temp-file "some/other/testfile1" (insert "test"))
+        (with-temp-file "some/other/testfile2" (insert "test"))
+        ))))
 
 (defun bash-completion_test-teardown-env (test-env-dir)
   "Deletes everything `bash-completion_test-setup-env' set up."
   (when test-env-dir
-    (if (>= emacs-major-version 24)
-        (delete-directory test-env-dir 'recursive)
-      (dired-delete-file test-env-dir 'always))))
+    (delete-directory test-env-dir 'recursive)))
 
 (defun bash-completion_test-equal-any-order (expected actual)
   "Compare a sorted list of string EXPECTED with ACTUAL.
@@ -230,12 +255,12 @@ across Emacs version."
    (should (equal "somefunction so"
                   (bash-completion_test-complete "somefunction so"))) ;
    ;; function returns nothing, -o default, so fallback to default 
-   (should (equal "someotherfunction some/"
-                   (bash-completion_test-complete "someotherfunction so")))
+   (should (equal "someotherfunction moretestfile "
+                   (bash-completion_test-complete "someotherfunction moret")))
    ;; wordbreak completion
-   (should (equal "export SOMEPATH=some/directory:some/other/"
+   (should (equal "export SOMEPATH=some/directory:some/testfile "
                   (bash-completion_test-complete
-                   "export SOMEPATH=some/directory:some/oth")))))
+                   "export SOMEPATH=some/directory:some/te")))))
 
 (ert-deftest bash-completion-integration-multiple-completions-test ()
   (bash-completion_test-multiple-completions-test ""))
@@ -249,10 +274,10 @@ across Emacs version."
    bashrc
    nil ; use-separate-process
    (should (bash-completion_test-equal-any-order
-            '("some/directory/" "some/other/")
-            (bash-completion_test-candidates "ls some/")))
-   (should (equal '("some/directory/") (bash-completion_test-candidates "ls some/d")))
-   (should (equal '("some/directory/") (bash-completion_test-candidates "ls some/di")))
+            '("some/other/testfile1" "some/other/testfile2")
+            (bash-completion_test-candidates "ls some/other/te")))
+   (should (equal '("some/testfile ") (bash-completion_test-candidates "ls some/te")))
+   (should (equal '("some/testfile ") (bash-completion_test-candidates "ls some/te")))
    (should (equal '() (bash-completion_test-candidates "ls some/do")))))
 
 (ert-deftest bash-completion-integration-nocomint-test ()
@@ -289,25 +314,82 @@ across Emacs version."
       ;; actually bash, the test could otherwise work just fine.
       (should (not (null (cdr (assq nil bash-completion-processes)))))))))
 
+(ert-deftest bash-completion-integration-notbash-anymore-test ()
+  (skip-unless bash-completion_test-notbash-prog)
+  (bash-completion_test-with-shell-harness
+   "" ; bashrc
+   nil ; use-separate-process
+   ;; initially completion works
+   (should (bash-completion_test-equal-any-order
+            '("testfile1" "testfile2")
+            (bash-completion_test-candidates "ls testf")))
+   ;; but then later on bash executes another command-line tool
+   (bash-completion_test-send-nowait bash-completion_test-notbash-prog)
+   (let ((bash-completion--debug-info nil))
+     (should-error (bash-completion_test-candidates "ls testf"))
+     (should (equal "short-timeout" (cdr (assq 'error bash-completion--debug-info)))))
+
+   ;; this is recoverable, once the other command-line tool exits,
+   ;; completion works again.
+   (bash-completion_test-send-nowait "\004") ;; C-d
+
+   (should (bash-completion_test-equal-any-order
+            '("testfile1" "testfile2")
+            (bash-completion_test-candidates "ls testf")))))
+
+(ert-deftest bash-completion-integration-bash-in-bash ()
+  (skip-unless bash-completion_test-notbash-prog)
+  (bash-completion_test-with-shell-harness
+   "" ; bashrc
+   nil ; use-separate-process
+   ;; initially completion works
+   (should (bash-completion_test-equal-any-order
+            '("testfile1" "testfile2")
+            (bash-completion_test-candidates "ls testf")))
+
+   ;; Start a bash subprocess. Note that Emacs normally adds
+   ;; --noediting, which isn't set here; it should work nevertheless.
+   (bash-completion_test-send bash-completion-prog)
+   (bash-completion_test-send "HISTFILE=/dev/null") ;; avoid history file
+   (bash-completion_test-send "function sometest { echo ok; }")
+   (should (bash-completion_test-equal-any-order
+            '("sometest ")
+            (bash-completion_test-candidates "somet")))
+   (should (bash-completion_test-equal-any-order
+            '("testfile1" "testfile2")
+            (bash-completion_test-candidates "ls testf")))
+   (bash-completion_test-send "exit")
+
+   ;; back in the main bash process, completion results are different,
+   ;; as sometest isn't defined.
+   (should (bash-completion_test-equal-any-order
+            '() (bash-completion_test-candidates "somet")))
+
+   (should (bash-completion_test-equal-any-order
+            '("testfile1" "testfile2")
+            (bash-completion_test-candidates "ls testf")))))
+
 (ert-deftest bash-completion-integration-space ()
+  ;; While completion generally works with Bash 4.0, 4.1, 4.2 and 4.3,
+  ;; bash-completion.el appending / to directories doesn't work.
   (bash-completion_test-with-shell-harness
    ""
    t ; bash-completion-use-separate-processes
    (bash-completion_test-test-spaces)))
 
 (ert-deftest bash-completion-integration-space-and-prog-completion ()
+  (skip-unless (and bash-completion_test-setup-completion
+                    (not (zerop (length bash-completion_test-setup-completion)))))
   ;; Recent version of bash completion define a completion for ls. This
   ;; test makes sure that it works.
-  (when (and bash-completion_test-setup-completion
-             (not (zerop (length bash-completion_test-setup-completion))))
-    (bash-completion_test-with-shell-harness
-     (concat "source " bash-completion_test-setup-completion "\n")
-     t ; bash-completion-use-separate-processes
-     (bash-completion_test-test-spaces))))
+  (bash-completion_test-with-shell-harness
+   (concat "source " bash-completion_test-setup-completion "\n")
+   t ; bash-completion-use-separate-processes
+   (bash-completion_test-test-spaces)))
   
 (defun bash-completion_test-test-spaces ()
    (make-directory "my dir1/my dir2" 'parents)
-   (with-temp-buffer (write-file "my dir1/other"))
+   (with-temp-file "my dir1/other" (insert "test"))
 
    (should (equal "ls my\\ dir1/" (bash-completion_test-complete "ls my")))
    (should (equal "ls my\\ dir1/my\\ dir2/" (bash-completion_test-complete "ls my\\ dir1/my")))
@@ -344,11 +426,10 @@ across Emacs version."
     "}\n"
     "complete -D -F _default\n")
    use-separate-process
-   (when (>= (bash-completion_test-bash-major-version) 4)
-     (should (equal "dosomething dummy "
-                    (bash-completion_test-complete "dosomething du")))
-     (should (equal "dosomethingelse du"
-                      (bash-completion_test-complete "dosomethingelse du"))))))
+   (should (equal "dosomething dummy "
+                  (bash-completion_test-complete "dosomething du")))
+   (should (equal "dosomethingelse du"
+                  (bash-completion_test-complete "dosomethingelse du")))))
 
 (ert-deftest bash-completion-integration-bash-4-compopt ()
   (bash-completion_test-with-shell-harness
@@ -374,26 +455,25 @@ across Emacs version."
     "complete -F _sometimes_nospace sometimes_nospace\n"
     "complete -F _sometimes_not_nospace -o nospace sometimes_not_nospace\n")
    t ; use-separate-process
-   (when (>= (bash-completion_test-bash-major-version) 4)
+   (should (equal
+            "sometimes_nospace dummy "
+            (bash-completion_test-complete "sometimes_nospace du")))
+   (should (equal
+            "sometimes_nospace dummyo"
+            (bash-completion_test-complete "sometimes_nospace dum")))
+   (should (equal
+            "sometimes_not_nospace dummy"
+            (bash-completion_test-complete "sometimes_not_nospace du")))
+   (should (equal
+            "sometimes_not_nospace dummyo "
+            (bash-completion_test-complete "sometimes_not_nospace dum")))
+   (let ((bash-completion-nospace t)) ;; never nospace
      (should (equal
-              "sometimes_nospace dummy "
+              "sometimes_nospace dummy"
               (bash-completion_test-complete "sometimes_nospace du")))
      (should (equal
-              "sometimes_nospace dummyo"
-              (bash-completion_test-complete "sometimes_nospace dum")))
-     (should (equal
-              "sometimes_not_nospace dummy"
-              (bash-completion_test-complete "sometimes_not_nospace du")))
-     (should (equal
-              "sometimes_not_nospace dummyo "
-              (bash-completion_test-complete "sometimes_not_nospace dum")))
-     (let ((bash-completion-nospace t)) ;; never nospace
-       (should (equal
-                "sometimes_nospace dummy"
-                (bash-completion_test-complete "sometimes_nospace du")))
-       (should (equal
-                "sometimes_not_nospace dummyo"
-                (bash-completion_test-complete "sometimes_not_nospace dum")))))))
+              "sometimes_not_nospace dummyo"
+              (bash-completion_test-complete "sometimes_not_nospace dum"))))))
 
 (ert-deftest bash-completion-integration-bash-4-substring-completion ()
   (bash-completion_test-with-shell-harness
@@ -430,6 +510,14 @@ across Emacs version."
    (should (equal "dummy 1 Yooo "
                   (bash-completion_test-complete "dummy 1 Y")))))
 
+(ert-deftest bash-completion-integration-tracing ()
+  (bash-completion_test-with-shell-harness
+   "set -x\n" ;; .bashrc
+   nil ;; use-separate-process
+   (should (bash-completion_test-equal-any-order
+            '("testfile1" "testfile2")
+            (bash-completion_test-candidates "ls te")))))
+
 (ert-deftest bash-completion-integration-refresh-test ()
   (bash-completion_test-with-shell-harness
    (concat ; .bashrc
@@ -445,6 +533,37 @@ across Emacs version."
    (should (equal "dummy 1 Yaaa "
                   (bash-completion_test-complete "dummy 1 Y")))))
 
+(ert-deftest bash-completion-integration-default ()
+  (bash-completion_test-with-shell-harness
+   (concat ; .bashrc
+    "function _ook {\n"
+    "  COMPREPLY=(ook)\n"
+    "}\n"
+    "function _bar {\n"
+    "  COMPREPLY=(bar)\n"
+    "}\n"
+    "function _baa {\n"
+    "  COMPREPLY=(baa)\n"
+    "}\n"
+    "complete -F _ook -D\n"
+    "complete -F _bar foo\n"
+    "complete -F _baa baa\n")
+   t ; use-separate-process
+   ;; using default (_ook)
+   (should (equal
+            "faa ook "
+            (bash-completion_test-complete "faa a")))
+   (should (equal
+            "fee ook "
+            (bash-completion_test-complete "fee a")))
+   ;; exceptions, to make sure non-default still works
+   (should (equal
+            "foo bar "
+            (bash-completion_test-complete "foo a")))
+   (should (equal
+            "baa baa "
+            (bash-completion_test-complete "baa a")))))
+
 (ert-deftest bash-completion-integration-case-insensitive-test ()
   (bash-completion_test-harness
    (concat ; .bashrc
@@ -452,36 +571,35 @@ across Emacs version."
    nil ; use-separate-process
    (with-temp-file "test-inputrc"
      (insert "set completion-ignore-case on\n"))
-   (make-directory "Uppercase")
-   (make-directory "Another Uppercase")
-   (make-directory "libs")
-   (make-directory "Library")
+   (with-temp-file "Uppercase" (insert "test"))
+   (with-temp-file "Another Uppercase" (insert "test"))
+   (with-temp-file "libs" (insert "test"))
+   (with-temp-file "library" (insert "test"))
    (bash-completion_test-with-shell
-    (when (>= (bash-completion_test-bash-major-version) 4)
-      ;; Case insensitive completion is done by compgen which, under
-      ;; bash 4, respects the case sensitivity settings set in
-      ;; .inputrc.
-      (should (equal "ls some/" (bash-completion_test-complete "ls so")))
-      (should (equal "ls some/" (bash-completion_test-complete "ls So")))
-      (should (equal "ls Uppercase/" (bash-completion_test-complete "ls Up")))
-      (should (equal "ls Uppercase/" (bash-completion_test-complete "ls up")))
-      
-      (should (equal "ls libs/" (bash-completion_test-complete "ls li")))
-      (should (equal "ls libs/" (bash-completion_test-complete "ls Li")))
+    ;; Case insensitive completion is done by compgen which, under
+    ;; bash 4, respects the case sensitivity settings set in
+    ;; .inputrc.
+    (should (equal "ls testfile1" (bash-completion_test-complete "ls testf")))
+    (should (equal "ls testfile1" (bash-completion_test-complete "ls TestF")))
+    (should (equal "ls Uppercase " (bash-completion_test-complete "ls Up")))
+    (should (equal "ls Uppercase " (bash-completion_test-complete "ls up")))
+    
+    (should (equal "ls libs" (bash-completion_test-complete "ls li")))
+    (should (equal "ls libs" (bash-completion_test-complete "ls Li")))
 
-      (should (equal "ls Another\\ Uppercase/" (bash-completion_test-complete "ls Ano")))
-      (should (equal "ls Another\\ Uppercase/" (bash-completion_test-complete "ls ano")))
-      (should (equal "ls \"Another Uppercase/" (bash-completion_test-complete "ls \"Ano")))
-      (should (equal "ls \"Another Uppercase/" (bash-completion_test-complete "ls \"ano")))
-      (should (equal "ls 'Another Uppercase/" (bash-completion_test-complete "ls 'Ano")))
-      (should (equal "ls 'Another Uppercase/" (bash-completion_test-complete "ls 'ano")))
+    (should (equal "ls Another\\ Uppercase " (bash-completion_test-complete "ls Ano")))
+    (should (equal "ls Another\\ Uppercase " (bash-completion_test-complete "ls ano")))
+    (should (equal "ls \"Another Uppercase\" " (bash-completion_test-complete "ls \"Ano")))
+    (should (equal "ls \"Another Uppercase\" " (bash-completion_test-complete "ls \"ano")))
+    (should (equal "ls 'Another Uppercase' " (bash-completion_test-complete "ls 'Ano")))
+    (should (equal "ls 'Another Uppercase' " (bash-completion_test-complete "ls 'ano")))
 
-      ;; When doing case-insensitive search, bash-completion.el cannot
-      ;; keep the exact same quotes, so it just puts the quote, if
-      ;; any, at the beginning, just after the tilde part.
-      (should (equal "ls \"Another Uppercase/" (bash-completion_test-complete "ls ano\"t")))
-      (should (equal "ls 'Another Uppercase/" (bash-completion_test-complete "ls ano't")))
-      (should (equal "ls ~/\"Another Uppercase/" (bash-completion_test-complete "ls ~/ano\"t")))))))      
+    ;; When doing case-insensitive search, bash-completion.el cannot
+    ;; keep the exact same quotes, so it just puts the quote, if
+    ;; any, at the beginning, just after the tilde part.
+    (should (equal "ls \"Another Uppercase\" " (bash-completion_test-complete "ls ano\"t")))
+    (should (equal "ls 'Another Uppercase' " (bash-completion_test-complete "ls ano't")))
+    (should (equal "ls ~/\"Another Uppercase\" " (bash-completion_test-complete "ls ~/ano\"t"))))))
 
 (ert-deftest bash-completion-integration-case-sensitive-test ()
   (bash-completion_test-harness
@@ -490,11 +608,11 @@ across Emacs version."
    nil ; use-separate-process
    (with-temp-file "test-inputrc"
      (insert "set completion-ignore-case off\n"))
-   (make-directory "Uppercase")
+   (with-temp-file "Uppercase" (insert "test"))
    (bash-completion_test-with-shell
-    (should (equal "ls some/" (bash-completion_test-complete "ls so")))
-    (should (equal "ls So" (bash-completion_test-complete "ls So")))
-    (should (equal "ls Uppercase/" (bash-completion_test-complete "ls Up")))
+    (should (equal "ls moretestfile " (bash-completion_test-complete "ls moret")))
+    (should (equal "ls Te" (bash-completion_test-complete "ls Te")))
+    (should (equal "ls Uppercase " (bash-completion_test-complete "ls Up")))
     (should (equal "ls up" (bash-completion_test-complete "ls up")))
     (should (not completion-ignore-case)))))
 
@@ -502,10 +620,10 @@ across Emacs version."
   (bash-completion_test-with-shell-harness
    ""
    nil ; use-separate-process
-   (should (equal "ls some/" (bash-completion_test-complete "ls so")))
-   (should (equal "ls ~/some/" (bash-completion_test-complete "ls ~/so")))
-   (should (equal "ls \"~/some/" (bash-completion_test-complete "ls \"~/so")))
-   (should (equal "ls '~/some/" (bash-completion_test-complete "ls '~/so")))))
+   (should (equal "ls moretestfile " (bash-completion_test-complete "ls morete")))
+   (should (equal "ls ~/moretestfile " (bash-completion_test-complete "ls ~/morete")))
+   (should (equal "ls \"~/moretestfile\" " (bash-completion_test-complete "ls \"~/morete")))
+   (should (equal "ls '~/moretestfile' " (bash-completion_test-complete "ls '~/morete")))))
 
 (ert-deftest bash-completion-integration-prompt-command ()
   "Tests PROMPT_COMMAND storage and recovery in single-process mode."
@@ -518,14 +636,15 @@ function _prompt {
 PROMPT_COMMAND=_prompt
 "
    nil ; use-separate-process
-   (bash-completion_test-send "ls -1 so" 'complete)
+   (bash-completion_test-send "ls -1 morete" 'complete)
    (bash-completion_test-send "tru" 'complete)
    (bash-completion_test-send "fals" 'complete)
+   ;; PROMPT_COMMAND is reset early on to avoid ever running it when
+   ;; doing completion.
    (should (equal
             (bash-completion_test-buffer-string)
-            "[0]:0 $ ls -1 some/
-directory
-other
+            "[0]:0 $ ls -1 moretestfile
+moretestfile
 [1]:0 $ true
 [2]:0 $ false
 [3]:1 $ "))))
@@ -535,14 +654,13 @@ other
   (bash-completion_test-with-shell-harness
    "PS1='$? $ '"
    nil ; use-separate-process
-   (bash-completion_test-send "ls -1 so" 'complete)
+   (bash-completion_test-send "ls -1 morete" 'complete)
    (bash-completion_test-send "tru" 'complete)
    (bash-completion_test-send "fals" 'complete)
    (should (equal
             (bash-completion_test-buffer-string)
-            "0 $ ls -1 some/
-directory
-other
+            "0 $ ls -1 moretestfile
+moretestfile
 0 $ true
 0 $ false
 1 $ "))))
@@ -552,17 +670,16 @@ other
   (bash-completion_test-with-shell-harness
    "PS1='$ '"
    nil ; use-separate-process
-   (bash-completion_test-send "ls -1 so" 'complete)
+   (bash-completion_test-send "ls -1 morete" 'complete)
    (bash-completion_test-send "tru" 'complete)
    (bash-completion_test-send "fals" 'complete)
-   (let ((history-start (line-beginning-position)))
-     (bash-completion_test-send "history")
+   (let ((history-start (bash-completion_test-send "history")))
      (untabify (point-min) (point-max))
      (delete-trailing-whitespace (point-min) (point-max))
      (should (equal
               (bash-completion_test-buffer-string history-start)
               "history
-    1  ls -1 some/
+    1  ls -1 moretestfile
     2  true
     3  false
     4  history
@@ -574,7 +691,7 @@ $ ")))))
    ""  ; .bashrc
    nil ; use-separate-process
    (let ((default-directory "/does-not-exist/"))
-     (should (equal "ls some/" (bash-completion_test-complete "ls so"))))))
+     (should (equal "ls moretestfile " (bash-completion_test-complete "ls moret"))))))
 
 (ert-deftest bash-completion-integration-caching ()
   "Make sure caching works and that completion is only executed once."
@@ -599,8 +716,8 @@ $ ")))))
    ""
    nil ; use-separate-process
    (let ((default-directory test-env-dir))
-     (make-directory "some/file" 'parents)
-     (make-directory "some/filled" 'parents))
+     (with-temp-file "some/file" (insert "test"))
+     (with-temp-file "some/filled" (insert "test")))
    (let ((compfunc-some (bash-completion_test-candidates "ls some/f" 'dynamic-table))
          (compfunc-one (bash-completion_test-candidates "ls some/fill" 'dynamic-table)))
 
@@ -610,45 +727,45 @@ $ ")))))
      
      ;; all-completion
      (should (bash-completion_test-equal-any-order
-              '("some/file/" "some/filled/") (funcall compfunc-some "some/" nil t)))
-     (should (equal '("some/filled/") (funcall compfunc-some "some/fill" nil t)))
+              '("some/file" "some/filled") (funcall compfunc-some "some/" nil t)))
+     (should (equal '("some/filled") (funcall compfunc-some "some/fill" nil t)))
      (should (equal nil (funcall compfunc-some "other" nil t)))
-     (should (equal '("some/filled/") (funcall compfunc-one "some/fill" nil t)))
+     (should (equal '("some/filled ") (funcall compfunc-one "some/fill" nil t)))
 
      ;; all-completion with predicate
-     (should (equal '("some/file/")
+     (should (equal '("some/file")
                     (funcall compfunc-some "some/"
-                             (lambda (c) (string= c "some/file/")) t)))
+                             (lambda (c) (string= c "some/file")) t)))
      (should (equal nil
                     (funcall compfunc-some "some/" (lambda (c) nil) t)))
      (should (bash-completion_test-equal-any-order
-              '("some/file/" "some/filled/")
+              '("some/file" "some/filled")
               (funcall compfunc-some "some/" (lambda (c) t) t)))
 
      ;; try-completion
-     (should (equal "some/filled/" (funcall compfunc-one "some/fill" nil nil)))
+     (should (equal "some/filled " (funcall compfunc-one "some/fill" nil nil)))
      (should (equal "some/fil" (funcall compfunc-some "some/" nil nil)))
-     (should (equal t (funcall compfunc-some "some/file/" nil nil)))
-     (should (equal t (funcall compfunc-one "some/filled/" nil nil)))
+     (should (equal t (funcall compfunc-some "some/file" nil nil)))
+     (should (equal t (funcall compfunc-one "some/filled " nil nil)))
 
      ;; try-completion with predicate
-     (should (equal "some/file/"
+     (should (equal "some/file"
                     (funcall compfunc-some "some/"
-                             (lambda (c) (string= c "some/file/")) nil)))
+                             (lambda (c) (string= c "some/file")) nil)))
      
      ;; test-completion
      (should (equal nil (funcall compfunc-some "some/" nil 'lambda)))
-     (should (equal t (funcall compfunc-some "some/file/" nil 'lambda)))
-     (should (equal t (funcall compfunc-some "some/filled/" nil 'lambda)))
+     (should (equal t (funcall compfunc-some "some/file" nil 'lambda)))
+     (should (equal t (funcall compfunc-some "some/filled" nil 'lambda)))
      (should (equal nil (funcall compfunc-one "some/fill" nil 'lambda)))
-     (should (equal t (funcall compfunc-one "some/filled/" nil 'lambda)))
+     (should (equal t (funcall compfunc-one "some/filled " nil 'lambda)))
 
      ;; test-completion with predicate
      (should (equal nil (funcall compfunc-some "some/" nil 'lambda)))
      (should (equal nil (funcall compfunc-some "some/file/"
-                                 (lambda (c) (string= c "some/filled/")) 'lambda)))
-     (should (equal t (funcall compfunc-some "some/filled/"
-                               (lambda (c) (string= c "some/filled/")) 'lambda)))
+                                 (lambda (c) (string= c "some/filled")) 'lambda)))
+     (should (equal t (funcall compfunc-some "some/filled"
+                               (lambda (c) (string= c "some/filled")) 'lambda)))
 
      ;; completion-boundaries (not supported)
      (should (equal nil (funcall compfunc-some "some/" nil '(boundaries . "/"))))
@@ -736,4 +853,51 @@ $ ")))))
      (should (equal nil (funcall compfunc-nonprefix "babeetai"
                                  (lambda (c) (equal "babeedai" c)) 'lambda))))))
 
+(ert-deftest bash-completion_test-issue-74 ()
+  (bash-completion_test-with-shell-harness
+   (concat ; .bashrc
+    "_mycmd() {\n"
+    "  readarray -t opts < <(printf \"foo\nbar\")\n"
+    "  COMPREPLY=($(compgen -W \"${opts[*]}\" -- \"${COMP_WORDS[$COMP_CWORD]}\"))\n"
+    "}\n"
+    "complete -F _mycmd mycmd\n")
+   nil ;; use-separate-process
+
+   (should (equal "mycmd bar "
+                  (bash-completion_test-complete "mycmd b")))))
+
+(ert-deftest bash-completion-integration-filenames-option ()
+  (bash-completion_test-with-shell-harness
+   (concat ; .bashrc
+    "function _dummy {\n"
+    "}\n"
+    "complete -F _dummy -o plusdir dummy\n")
+   nil
+   (should (equal
+            "dummy some/other/"
+            (bash-completion_test-complete "dummy some/ot")))))
+
+(ert-deftest bash-completion-integration-plusdir-option ()
+  (bash-completion_test-with-shell-harness
+   (concat ; .bashrc
+    "function _dummy {\n"
+    "}\n"
+    "complete -F _dummy -o filenames dummy\n")
+   nil
+   (should (equal
+            "dummy moretestfile "
+            (bash-completion_test-complete "dummy moret")))))
+
+(ert-deftest bash-completion-integration-new-compopt-options ()
+  (bash-completion_test-with-shell-harness
+   (concat ; .bashrc
+    "function _dummy {\n"
+    "    compopt -o default -o bashdefault\n"
+    "}\n"
+    "complete -F _dummy dummy\n")
+   nil
+   (should (equal
+            "dummy moretestfile "
+            (bash-completion_test-complete "dummy moret")))))
+
 ;;; bash-completion-integration-test.el ends here
